#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from subprocess import Popen, PIPE
from optparse import OptionParser, OptionGroup
import os, os.path, re, signal, string, sys
import threading, time, urllib, fcntl, traceback

# GUI stuff
try:
    import pygtk
    import gtk, gobject, webbrowser
    gtk.gdk.threads_init()
    pygtk_available = True
except:
    pygtk_available = False


app_license = """
foo2svi - convert any video to SVI format
Copyright (C) 2008  nikosapi (me@nikosapi.org)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
app_name = 'foo2svi'
app_version = '1.1'
app_copyright = '© 2008 nikosapi'
app_website = 'http://nikosapi.org/wiki/index.php/Foo2svi'

icon_dirs = [ '/usr/share/pixmaps', '/usr/local/share/pixmaps', '.' ]
# Note: Icon comes from the Tango project (because I have no artistic talent).
# Source filename: face-devilish.svg
# Source URL: http://tinyurl.com/face-devilish-svg
icon_name = 'foo2svi.svg'

class foo2svi_gui(object):
    def __init__( self, video=None ):
        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.icon = find_icon( icon_name, icon_dirs )
        if self.icon is not None:
            self.window.set_icon_from_file( self.icon )
        self.window.connect('destroy', self.on_destroy_event)
        self.window.resize(350,130)
        self.window.set_geometry_hints(None, min_width=340, min_height=110,
            max_width=400, max_height=150)
        self.window.set_title(app_name)

        main_vbox = gtk.VBox(False, 2)
        main_vbox.set_border_width(0)
        self.window.add(main_vbox)
        main_vbox.show()

        menubar = gtk.MenuBar()
        main_vbox.pack_start(menubar, False, False, 0)
        menubar.show()

        file_menu = gtk.Menu()
        action_group = gtk.ActionGroup('menubar')

        self.openfile_menu_item = gtk.MenuItem('Open Video File...')
        self.openfile_menu_item.connect("activate",
            self.on_open_video_file_activate)
        file_menu.append(self.openfile_menu_item)
        self.openfile_menu_item.show()

        menu_item = gtk.MenuItem('Preferences')
        menu_item.connect("activate", self.on_preferences_activate)
        file_menu.append(menu_item)
        menu_item.show()

        menu_item = gtk.MenuItem('Close')
        menu_item.connect("activate", self.on_destroy_event)
        file_menu.append(menu_item)
        menu_item.show()

        file_action = gtk.Action('File', '_File', None, None)
        action_group.add_action(file_action)
        file_menuitem = file_action.create_menu_item()
        file_menuitem.show()
        file_menuitem.set_submenu(file_menu)

        about_menu = gtk.Menu()

        menu_item = gtk.MenuItem('About')
        menu_item.connect("activate", self.on_about_activate)
        about_menu.append(menu_item)
        menu_item.show()

        gtk.about_dialog_set_url_hook( lambda dlg, link, user_data=None:
            webbrowser.open(link) )

        help_action = gtk.Action('Help', '_Help', None, None)
        action_group.add_action(help_action)
        help_menuitem = help_action.create_menu_item()
        help_menuitem.show()
        help_menuitem.set_submenu(about_menu)

        menubar.append(file_menuitem)
        menubar.append(help_menuitem)

        table = gtk.Table(2, 2)
        table.set_border_width(10)
        main_vbox.add(table)
        table.show()

        self.progress = gtk.ProgressBar()
        self.welcome_message = 'Welcome to %s, load a video to continue' % (
            app_name )
        self.progress.set_text(self.welcome_message)
        table.attach(self.progress, 0,2,0,1, ypadding=3)
        # set up drag and drop for the progress bar
        self.enable_dnd( self.progress )
        self.progress.show()

        self.button_open = gtk.Button(stock=gtk.STOCK_OPEN)
        label = self.button_open.get_children()[0]
        label.get_children()[0].get_children()[1].set_label('Load Video')
        self.button_open.connect("clicked", self.on_open_video_file_activate)
        table.attach(self.button_open, 1,2,1,2,
            xoptions=gtk.SHRINK|gtk.FILL,  ypadding=3)

        self.button_cancel = gtk.Button(stock=gtk.STOCK_CANCEL)
        self.button_cancel.connect("clicked", self.on_cancel_encode_activate)
        table.attach(self.button_cancel, 1,2,1,2,
            xoptions=gtk.SHRINK|gtk.FILL, ypadding=3)

        self.button_convert = gtk.Button(stock=gtk.STOCK_CONVERT)
        self.button_convert.connect("clicked", self.on_convert_activate)
        table.attach(self.button_convert, 1,2,1,2,
            xoptions=gtk.SHRINK|gtk.FILL, ypadding=3)
        
        self.button_open.show()
        self.window.show()

        if video is not None and os.path.isfile(video):
            self.video_file = video
            self.load_video(video)
        else:
            self.video_file = ''

        # Get the default settings
        self.settings = SettingsManager().load()

        # some stuff that's generated when the video is loaded
        self.output_file = ''
        self.video_info = []
        self.video_crop = []
        self.video_size = []

        # The Popen object returned from run_command()
        self.mencoder_process = None
        self.encoding_cancelled = False

    def on_destroy_event( self, widget, *arg ):
        self.settings.save()
        gtk.main_quit()

    def on_dnd_drag_data_received_event( self, widget, drag_context, x, y,
        selection_data, info, time ):
        data = selection_data.data.strip()
        if data and data.startswith('file://'):
            if len(data.split('\n')) == 1:
                filename = os.path.expanduser(urllib.unquote(data[7:]))
                if os.path.isfile(filename):
                    self.load_video(filename)
            elif len(data.split('\n')) > 1:
                self.progressbar_callback( 0, 'Please choose only one file.' )
                self.show_hide_buttons(self.button_open)
                gobject.timeout_add( 3000, self.progressbar_callback, 0,
                    self.welcome_message )
        drag_context.finish( True, False, time )

    def on_open_video_file_activate( self, widget, *arg ):
        videos_dir = os.path.join(os.path.expanduser('~'),'Videos')
        if not os.path.isdir(videos_dir):
            videos_dir = os.path.expanduser('~')

        filename = file_chooser( 'Choose a Video File to Convert',
            choose_file=True, filters=[('Video Files', False, 'video/*'),
            ('All Files', True, '*')], starting_dir=videos_dir )

        if filename is not None and os.path.isfile(filename):
            self.load_video(filename)

    def on_cancel_encode_activate( self, widget, *args ):
        self.encoding_cancelled = True

    def on_convert_activate( self, widget, *args ):
        self.encode_video()

    def on_preferences_activate( self, widget, *args ):
        preferences_gui( self, self.settings )

    def on_about_activate( self, widget, *args ):
    	about = gtk.AboutDialog()
    	if self.icon is not None:
    	    about.set_logo(gtk.gdk.pixbuf_new_from_file( self.icon ))
    	about.set_transient_for(self.window)
    	about.set_name(app_name)
    	about.set_version(app_version)
    	about.set_copyright(app_copyright)
    	about.set_license(app_license.strip())
    	about.set_website(app_website)
    	about.connect( 'response', lambda about, response: about.destroy())
    	about.run()

    def enable_dnd( self, widget, enable=True ):
        if enable:
            flags = gtk.DEST_DEFAULT_ALL
            targets = [ ('text/plain', 0, 0), ]
            actions = gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_MOVE
            widget.drag_dest_set( flags, targets, actions )
            widget.connect( 'drag_data_received',
                self.on_dnd_drag_data_received_event )
        else:
            widget.drag_dest_unset()

    def show_hide_buttons(self, show, sensitive=True):
        buttons = (self.button_open, self.button_convert, self.button_cancel)
        for button in buttons:
            if button == show:
                button.show()
                button.set_sensitive(sensitive)
            else:
                button.hide()

        if show == self.button_open:
            self.openfile_menu_item.set_sensitive( sensitive )
            self.enable_dnd( self.progress, sensitive )
        else:
            self.openfile_menu_item.set_sensitive( True )
            self.enable_dnd( self.progress,
                self.button_convert.flags() & gtk.VISIBLE )

    def progressbar_callback(self, fraction, text, tooltip=None):
        gobject.idle_add( self.progress.set_fraction, fraction )
        gobject.idle_add( self.progress.set_text, text )
        if tooltip is not None:
            gobject.idle_add( self.progress.set_tooltip_text, tooltip )
        else:
            gobject.idle_add( self.progress.set_tooltip_text, '' )
            try:
                gobject.idle_add( self.progress.set_has_tooltip, False )
            except:
                # Work-around for pygtk bug #541296
                gobject.idle_add( gtk.Widget.set_tooltip_text, self.progress,
                    None )
                log('Pygtk set_has_tooltip() is broken on your system :(', 
                    sender=self )

    def load_finished_callback( self, error=False ):
        if not error:
            self.show_hide_buttons(self.button_convert)
        else:
            self.show_hide_buttons(self.button_open)

    def load_video_thread(self, video, callback, finished_callback):
        self.video_file = video
        # sleep a bit just to let the user know what's going on
        sleep = 0.5
        callback(0, 'Getting Video Details...')
        time.sleep(sleep)
        x,y,length,fps = self.video_info = get_video_info(MPLAYER, video)
        if None in self.video_info:
            callback(0, 'Error Loading Video :(')
            finished_callback(error=True)
            return
        cx,cy,shift_x,shift_y = self.video_crop = (x,y,0,0)
        if self.settings.cropdetect:
            callback(1/3, 'Detecting Black Bars...')
            time.sleep(sleep)
            cx,cy,shift_x,shift_y = self.video_crop = detect_crop( MPLAYER,
                video, length, self.settings.cropdetect_limit )
        callback(2/3, 'Calculating Final Video size...')
        time.sleep(sleep)
        nx,ny = self.video_size = resize_video((cx,cy), ( 
            self.settings.constraint_x, self.settings.constraint_y ))
        callback(1, 'Video Loaded.')
        directory,video = os.path.split(video)
        self.output_file = os.path.join( os.path.expanduser(
            self.settings.save_dir),
            os.path.splitext(video)[0] + '.svi' )
        finished_callback()

    def load_video(self, video):
        self.show_hide_buttons(self.button_open, False)
        threading.Thread( target=self.load_video_thread, args=( video, 
            self.progressbar_callback, self.load_finished_callback ) ).start()

    def encode_video(self):
        self.progressbar_callback(0, 'About to encode...')
        self.show_hide_buttons(self.button_cancel)
        self.openfile_menu_item.set_sensitive(False)
        self.encoding_cancelled = False
        self.mencoder_process = encode_video_proc( MENCODER, self.video_file,
            self.output_file, self.video_size, self.video_crop, 
            self.settings.quantizer,self.settings.audio_bitrate )
        gobject.timeout_add( 50, self.encode_run_callback )

    def encode_run_callback(self):
        return encode_run_callback( self.mencoder_process,
            self.encoding_cancelled, self.encode_video_callback )

    def encode_video_callback(self, enc_status, returncode):
        ( pos, frame, percent, fps, trem, size, sync, unk ) = enc_status
        ( x, y, length, vfps ) = self.video_info
        r_percent, r_rtime = percent_and_rtime( frame, length*vfps, fps )
        string = 'Time Remaining: %s' % fancy_time_formater(r_rtime)
        tooltip_text = 'Estimated Size: %dMB\nEncoder Speed: %.1ffps' % (
            size, fps )
        self.progressbar_callback(r_percent, string, tooltip_text)

        if returncode is not None:
            self.show_hide_buttons(self.button_open)
            if not returncode:
                self.progressbar_callback(1, 'Encoding Complete!')
            elif self.encoding_cancelled:
                self.progressbar_callback(0, 'Encoding Cancelled!')
            else:
                log( 'Encoder returncode: %d' % returncode, sender=self )
                self.progressbar_callback(0, 'Error encoding :(')


def fancy_time_formater(time_in_seconds):
    s = max( 0, int( time_in_seconds ))
    if s == 1: 
        out = '1 Second'
    elif s < 60:
        out = '%d Seconds' % s
    elif s < 60*60:
        out = '%d Minutes' % ( s/60 )
    else:
        out = 'unknown'
    return out

def find_icon( icon_name, dirs ):
    for dir in dirs:
        icon = os.path.join( dir, icon_name )
        if os.path.isfile( icon ):
            return icon


class preferences_gui(object):
    def __init__(self, main_gui, settings):
        self.settings = settings

        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        if main_gui.icon is not None:
            self.window.set_icon_from_file(main_gui.icon)
        self.window.set_transient_for(main_gui.window)
        self.window.connect("destroy", self.on_destroy_event)
        self.window.resize(275,220)
        self.window.set_geometry_hints(None, min_width=260, min_height=210,
            max_width=400, max_height=245)
        self.window.set_title(app_name + ' preferences')

        main_vbox = gtk.VBox(False, 2)
        main_vbox.set_border_width(5)
        self.window.add(main_vbox)

        table = gtk.Table(8,3)
        table.set_border_width(5)
        main_vbox.pack_start(table)

        label = gtk.Label()
        label.set_markup('<b>Encoder Quality Setting</b>')
        label.set_alignment(0, 0.5)
        table.attach( label, 0,3,0,1 )

        # minimum quant == 15 or the video will most definitely look like shit
        self.quant_slider_adj = gtk.Adjustment( 5, 1, 15 )
        quant_slider = gtk.HScale(self.quant_slider_adj)
        quant_slider.set_draw_value(False)
        self.quant_slider_adj.set_value(self.settings.quantizer)
        table.attach(quant_slider, 1,2,1,2)

        label = gtk.Label('Quality')
        table.attach( label, 0,1,1,2, xoptions=0, xpadding=1 )

        label = gtk.Label('Speed')
        table.attach( label, 2,3,1,2, xoptions=0, xpadding=1 )

        label = gtk.Label()
        label.set_markup('<b>Automatic Crop Detection</b>')
        label.set_alignment(0, 0.5)
        table.attach( label, 0,3,2,3 )

        self.btn_cropdetect = gtk.CheckButton('Remove black bars from video')
        self.btn_cropdetect.set_active(self.settings.cropdetect)
        table.attach( self.btn_cropdetect, 0,3,3,4 )

        label = gtk.Label()
        label.set_markup('<b>Target Video Size</b>')
        label.set_alignment(0, 0.5)
        table.attach( label, 0,3,4,5 )

        size_hbox = gtk.HBox( False )
        table.attach( size_hbox, 0,3,5,6 )

        label = gtk.Label('Constraint in pixels (WxH):')
        size_hbox.pack_start( label )

        self.size_x = gtk.Entry(max=4)
        self.size_x.set_width_chars(4)
        self.size_x.set_text(str(self.settings.constraint_x))
        size_hbox.pack_end( self.size_x )

        label = gtk.Label('x')
        size_hbox.pack_end( label )

        self.size_y = gtk.Entry(max=4)
        self.size_y.set_width_chars(4)
        self.size_y.set_text(str(self.settings.constraint_y))
        size_hbox.pack_end( self.size_y )

        size_hbox.reorder_child(self.size_x, 3)
        size_hbox.reorder_child(self.size_y, 1)

        label = gtk.Label()
        label.set_markup('<b>Default Save Directory</b>')
        label.set_alignment(0, 0.5)
        table.attach( label, 0,3,6,7 )

        btn_savedir = gtk.Button(stock=gtk.STOCK_OPEN)
        label = btn_savedir.get_children()[0]
        self.savedir_label = label.get_children()[0].get_children()[1]
        self.savedir_label.set_label( self.settings.save_dir.replace(
            os.path.expanduser('~'), '~', 1) )
        btn_savedir.connect( 'clicked', self.on_btn_savedir_clicked )
        table.attach( btn_savedir, 0,3,7,8 )

        main_vbox.pack_start( gtk.HSeparator() )

        save_hbox = gtk.HBox( False )
        main_vbox.pack_end(save_hbox)

        btn = gtk.Button(stock=gtk.STOCK_SAVE)
        btn.connect( 'clicked', self.on_btn_save_clicked )
        save_hbox.pack_end(btn, expand=False, fill=False, padding=3)

        btn = gtk.Button(stock=gtk.STOCK_CANCEL)
        btn.connect( 'clicked', self.on_destroy_event )
        save_hbox.pack_end(btn, expand=False, fill=False, padding=3 )

        self.window.show_all()

    def on_destroy_event(self, widget, *args):
        self.settings.load()
        self.window.destroy()

    def on_btn_savedir_clicked( self, widget, *args ):
        self.settings.save_dir = file_chooser( 
            'Choose The Default Video Save Directory', choose_file=False,
            starting_dir=os.path.expanduser(self.settings.save_dir) )

        self.savedir_label.set_label( self.settings.save_dir.replace(
            os.path.expanduser('~'), '~', 1) )

    def on_btn_save_clicked(self, widget, *args):
        self.settings.quantizer = int(self.quant_slider_adj.get_value())
        self.settings.cropdetect = self.btn_cropdetect.get_active()
        try:
            self.settings.constraint_x = int(self.size_x.get_text())
            self.settings.constraint_y = int(self.size_y.get_text())
        except ValueError:
            log('Bad constraints, not saving...', sender=self)
        self.settings.save()
        self.window.destroy()


class SettingsManager(dict):
    default_settings = {
        'quantizer' : ( int, 5 ),
        'save_dir'  : ( str, '~/Desktop' ), 
        'cropdetect': ( bool, True ),
        'cropdetect_limit' : ( int, 100 ),
        'constraint_x' : ( int, 470 ),
        'constraint_y' : ( int, 272 ),
        'audio_bitrate' : ( int, 96 ), }

    def __init__(self, settings_file='~/.config/foo2svi.conf' ):
        self.__settings_file = os.path.expanduser(settings_file)
        dict.__init__(self)

    def get_setting(self, setting_name ):
        if self.default_settings.has_key(setting_name):
            return self[setting_name]

    def set_setting(self, setting_name, value ):
        if self.default_settings.has_key(setting_name):
            self[setting_name] = value

    def load(self):
        for setting,(type,value) in self.default_settings.iteritems():
            self[setting] = value
        if os.path.isfile(self.__settings_file):
            self.__load_settings()
        return self

    def __load_settings(self):
        bool_lookup = { 'True':True, 'False': False }
        try:
            f = open( self.__settings_file, 'r' )
            entries = [ i.strip() for i in f.readlines() ]
            f.close()
        except:
            entries = []
        for entry in entries:
            if entry.find(':') != -1:
                setting,value = entry.split(':',1)
                if setting in self.default_settings:
                    type,v = self.default_settings[setting]
                    if type == bool:
                        self[setting] = bool_lookup.get(value.strip(), False)
                    else:
                        self[setting] = type(value)

    def save(self):
        if not os.path.isfile(self.__settings_file):
            conf_dir = os.path.dirname(self.__settings_file)
            if os.path.isfile(conf_dir):
                print 'Error: Conf dir is a file.'
                return
            if not os.path.isdir(conf_dir):
                os.makedirs(conf_dir)
        try:
            f = open( self.__settings_file, 'w' )
        except IOError:
            print 'IOError, check your permissions.'
            return

        for setting in self.default_settings:
            if self.has_key(setting):
                f.write('%s:%s\n' % (setting, self[setting]))
        f.close()

    def merge(self, settings_manager_object):
        """ merge two SettingsManager objects together """
        for setting in self.default_settings:
            self[setting] = settings_manager_object[setting]

    def __getattr__(self, name):
        rtn = self.get_setting(name)
        if rtn is not None:
            return rtn
        else:
            raise AttributeError

    def __setattr__(self, name, value):
        if name in self.default_settings:
            self[name] = value
        else:
            object.__setattr__( self, name, value)


def file_chooser( title, starting_dir=None, choose_file=True, filters=None ):
    if choose_file:
        f_action = gtk.FILE_CHOOSER_ACTION_OPEN
    else:
        f_action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER

    chooser = gtk.FileChooserDialog( title=title, action=f_action, buttons=(
    gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK) )

    if filters is not None:
        for name,is_pattern,pattern in filters:
            file_filter = gtk.FileFilter()
            file_filter.set_name(name)
            if is_pattern:
                file_filter.add_pattern(pattern)
            else:
                file_filter.add_mime_type(pattern)
            chooser.add_filter(file_filter)

    if starting_dir is not None:
        chooser.set_current_folder(starting_dir)

    exit = chooser.run()
    filename = chooser.get_filename()
    chooser.destroy()
    return filename if exit == gtk.RESPONSE_OK else None


def find_command( command ):
    """
    Searches the system's PATH for a specific command that is
    executable by the user. Returns the first occurence of an
    executable binary in the PATH, or None if the command is 
    not available. (from gPodder's util.py)
    """

    if 'PATH' not in os.environ:
        return None

    for path in os.environ['PATH'].split( os.pathsep):
        command_file = os.path.join( path, command)
        if os.path.isfile(command_file) and os.access(command_file, os.X_OK):
            return command_file
        
    return None

def run_command( *args ):
    return Popen( args, stdout=PIPE, stdin=PIPE, stderr=PIPE )

def get_output( *args ):
    p = run_command( *args )
    p.wait()
    return p.stdout.read()

def get_video_info( mplayer, video ):
    ( w, h ) = ( 0, 0 )
    midentify = get_output( mplayer, '-frames', '0', '-identify', video )
    w, h, l, fps = [ None ]*4
    for line in midentify.split('\n'):
        if line.startswith('ID_VIDEO_HEIGHT'):
            h = int(line.split('=')[1])
        elif line.startswith('ID_VIDEO_WIDTH'):
            w = int(line.split('=')[1])
        elif line.startswith('ID_LENGTH'):
            l = int(float(line.split('=')[1]))
        elif line.startswith('ID_VIDEO_FPS'):
            fps = int(float(line.split('=')[1]))
    return [ w, h, l, fps ]

def resize_video( dimensions, constraints ):
    """ get new x/y quardinates that fit into <constraints>
        while respecting the aspect ratio """
    x,y = dimensions
    max_x,max_y = constraints
    aspect = x/y
    if aspect > max_x/max_y:
        new_x = max_x
        new_y = new_x/aspect
        new_y = int(new_y) if not int(new_y) % 2 else int(new_y) + 1
    else:
        new_y = max_y
        new_x = new_y*aspect
        new_x = int(new_x) if not int(new_x) % 2 else int(new_x) + 1
    return [ new_x, new_y ]

def detect_crop( mplayer, video, length = 180, cropdetect_limit=100 ):
    dim = []
    mplayer_args = ( '-vf cropdetect=limit:%s -nosound -vo null -frames 15' % 
        cropdetect_limit ) + ' -ss %d'

    for ss in range( int(length/4), length, int(length/4) ):
        crop = get_output( mplayer, video, *(mplayer_args % ss).split() )
        for line in crop.split('\n'):
            search = re.search('\(-vf crop=(([0-9]+:){2}[0-9:-]+)\)', line)
            if search:
                dim.append( search.group(1) )

    counter = {}
    for crop in dim:
        counter[crop] = counter.get(crop, 0) + 1

    return [ int(i) for i in max(
        [(value, key) for (key, value) in counter.items()])[1].split(':') ]

def encode_video_cli( mencoder, v_in, v_out, (dim_x, dim_y), crop, quant,
    audio_bitrate, callback, *cb_args ):
    mencoder_proc = encode_video_proc( mencoder, v_in, v_out, (dim_x, dim_y),
        crop, quant, audio_bitrate )
    while mencoder_proc.poll() is None:
        encode_run_callback( mencoder_proc, False, callback, *cb_args )
        time.sleep(0.05)

def encode_run_callback( mencoder_process, encoding_cancelled, callback,
    *callback_args ):

    # damn, that's a nasty regex :P
    regex = string.join([ 'Pos: +([^s]+)s +([0-9]+)f +[(]([0-9 ]+)%[)]',
        '+([0-9.]+)fps Trem: +([0-9]+)min +([0-9]+)mb +A-V:([0-9.]+)',
        '+[[]([0-9:]+)[]]' ])
    regex = re.compile(regex)

    if encoding_cancelled:
        os.kill( mencoder_process.pid, signal.SIGTERM )

    if mencoder_process.poll() is None:

        ( pos, frame, percent, fps, trem, size, sync, unk ) = [ None, ]*8
        try:
            ( pos, frame, percent, fps, trem, size, sync, unk
            ) = regex.search(mencoder_process.stdout.read()).groups()

            pos, frame, percent  = float(pos), int(frame), int(percent)
            fps, trem = float(fps), int(trem)
            size, sync = int(size), float(sync)
        except:
            #print '***The following traceback is probably nothing serious***'
            #traceback.print_exc(file=sys.stdout)
            #print '---------------------------------------------------------'
            return True

        if None not in [ percent, fps, trem, size ]:
            callback( ( pos, frame, percent, fps, trem, size, sync, unk ),
                mencoder_process.returncode, *callback_args )
        return True
    else:        
        # call one last time to ensure the callback gets the returncode
        callback(  [0,]*7 + [''], mencoder_process.returncode, *callback_args )
        return False  # this makes gobject.timeout_add stop running

def encode_video_proc( mencoder, v_in, v_out, (dim_x, dim_y), crop, quant,
    audio_bitrate ):

    quant = quant if type(quant) == int else 5
    mencoder_args = [ ('-oac mp3lame -ovc xvid -lameopts cbr:br=%d ' +
        '-xvidencopts fixed_quant=%d:vhq=1:max_bframes=0 -of avi') % (
        audio_bitrate, quant ) ]
    
    if [dim_x, dim_y] == crop[:2]:
        mencoder_args.append('-vf scale=%d:%d' % (dim_x, dim_y))
    else:
        mencoder_args.append('-vf crop=%d:%d:%d:%d,scale=%d:%d' % tuple( 
            list(crop) + [dim_x, dim_y] ))

    mencoder = [ mencoder] + string.join(mencoder_args).split() + [ '-o',
        v_out, v_in ]
    log( 'Mencoder command:', *mencoder )

    mencoder_process = run_command( *mencoder )
    fcntl.fcntl( mencoder_process.stdout, fcntl.F_SETFL, os.O_NONBLOCK )
    return mencoder_process

def cli_encode_callback( enc_status, returncode ):
    """ Rely on mencoder's stats """
    ( pos, frame, percent, fps, trem, size, sync, unk ) = enc_status
    sys.stdout.write(('\r% 2d%% complete @ %.2ffps (%d min remaining, est.' +
        'size: %dMB)         ') % ( percent, fps, trem, size ))
    sys.stdout.flush()

def better_cli_encode_callback( enc_status, returncode, total_frames ):
    """ Rely on our own stats (more accurate) """
    ( pos, frame, percent, fps, trem, size, sync, unk ) = enc_status
    r_percent, r_trem = percent_and_rtime( frame, total_frames, fps )
    sys.stdout.write(('\r% 2d%% complete @ %.2ffps (%.1f min remaining, ' + 
        'est. size: %dMB)       ') % ( r_percent*100, fps, r_trem/60, size ))
    sys.stdout.flush()

def percent_and_rtime( frame, total_frames, fps ):
    r_percent = min(1, 
        float( frame / total_frames )) if total_frames != 0 else 0
    r_trem = max(0, int( (total_frames - frame) / fps )) if fps != 0 else 0
    return r_percent, r_trem

def log( *msg, **kwargs ):
    global verbose

    if not verbose:
        return

    sender = kwargs.get( 'sender', None )
    if sender is not None and type(sender) != str:
        sender = sender.__class__.__name__
    msg = '[%s] %s' % (sender if sender is not None else 'unknown',
        string.join(msg) )
    print msg

def optparser():
    usage = 'usage: %prog [options] input_video'
    parser = OptionParser(usage=usage, version='%prog '+'%s' % app_version)
    parser.add_option('-g', '--gui-mode', action='store_true',
        dest='gui_mode', default=False, help='Enable the GUI interface')
    parser.add_option('-c', '--cli-mode', action='store_true',
        dest='cli_mode', default=False, help='Enable the CLI interface')
    parser.add_option('-v', '--verbose', action='store_true', dest='verbose',
        default=False, help='Enable verbose logging')
    group = OptionGroup(parser, 'CLI Mode Options')
    group.add_option( '-o', '--output', dest='dest_file',
        help=('The destination file (by default file is placed in current' +
        ' directory)') )
    group.add_option( '-q', '--quantizer', type='int', dest='quant',
        default=5, help='Quantizer to use when encoding [Default: %default]')
    group.add_option( '--disable-bar-detection', action='store_true', 
        dest='disable_back_bar', default=False, 
        help='Disable auto-black bar detection')
    group.add_option( '-l', '--cropdetect-limit', dest='cropdetect_limit',
        type='int', default=100, help='mplayer cropdetect:limit setting ' +
        '[Default: %default]' )
    group.add_option( '-b', '--audio-bitrate', dest='audio_bitrate',
        type='int', default=96, help='Audio bitrate [Default: %default kb/s]' )
    group.add_option( '-d', '--constraint', dest='constraint',
        default='470x272', help=('Set the dimesions that the video must ' +  
            'fit into [Default: %default]' ))
    parser.add_option_group(group)
    return parser, parser.parse_args()

verbose = False
def main():
    global verbose
    parser, ( opts, args ) = optparser()
    video = args[0] if args else None
    verbose = opts.verbose

    if None in [ MPLAYER, MENCODER ]:
        print 'mplayer or mencoder are not available in your $PATH'
        sys.exit(1)

    if opts.gui_mode:
        if pygtk_available:
            foo2svi_gui(video)
            gtk.main()
            sys.exit(0)
        else:
            log('Error importing pygtk/gtk module, gui mode not available.')
    elif opts.cli_mode:
        if video is None:
            parser.error('Please specify an input video file.')

        player_dim = re.match('([0-9]+)x([0-9]+)', opts.constraint)
        if player_dim is not None:
            player_dim = [ int(i) for i in player_dim.groups() ]
        else:
            parser.error('[%s] is not a valid set of dimesions.' %
                opts.constraint)

        if opts.dest_file:
            output_video = opts.dest_file
        else:
            directory,video_file = os.path.split(video)
            output_video = os.path.join( directory, os.path.splitext(
                video_file)[0] + '.svi' )

        # get the original size, length and fps
        x,y,length,fps = get_video_info(MPLAYER, video)
        print 'Detected %dx%d video, checking for black bars...' % (x,y)

        # check if there are black bars
        cx,cy,shift_x,shift_y = x,y,0,0
        if not opts.disable_back_bar:
            cx,cy,shift_x,shift_y = detect_crop(MPLAYER, video, length, 
                opts.cropdetect_limit )

        # calculate the new video dimensions
        nx,ny = resize_video((cx,cy),player_dim)
        if [x,y] != [cx,cy]:
            print 'Black bars detected, cropping to %dx%d' % (cx,cy)
        print 'Resizing video to %dx%d' % (nx,ny)

        encode_video_cli( MENCODER, video, output_video, (nx,ny),
            (cx,cy,shift_x,shift_y), opts.quant, opts.audio_bitrate,
            better_cli_encode_callback, length*fps )
        print
    else:
        parser.print_help()

# Override these if you need to use a custom mplayer/mencoder
MENCODER = find_command('mencoder')
MPLAYER = find_command('mplayer')

if __name__ == "__main__":
    main()

